/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

gl_list_t mib_file_list;

struct slb_server *srv_head, *srv_tail;
size_t srv_count;

void
register_server(struct slb_server *srv)
{
	if (srv_tail)
		srv_tail->next = srv;
	else
		srv_head = srv;
	srv_tail = srv;
	srv_count++;
}


int
assert_string_arg(grecs_locus_t *locus,
		  enum grecs_callback_command cmd,
		  const grecs_value_t *value)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected scalar value as a tag"));
		return 1;
	}
	return 0;
}


static int
cb_mib_directory(enum grecs_callback_command cmd,
		 grecs_locus_t *locus,
		 void *varptr,
		 grecs_value_t *value,
		 void *cb_data)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected string value"));
		return 1;
	}
	add_mibdir(value->v.string);
	return 0;
}

int
cb_define_expression(enum grecs_callback_command cmd,
		     grecs_locus_t *locus,
		     void *varptr,
		     grecs_value_t *value,
		     void *cb_data)
{
	struct slb_expression *expr;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	if (compile_expression(value->v.arg.v[1].v.string, locus, &expr))
		return 1;
	expr->ex_name = value->v.arg.v[0].v.string;
	
	return register_expression(expr);
}

int
cb_server_expression(enum grecs_callback_command cmd,
		     grecs_locus_t *locus,
		     void *varptr,
		     grecs_value_t *value,
		     void *cb_data)
{
	struct slb_expression *expr, **pexpr = varptr;

	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected single string as argument"));
		return 1;
	}

	if (value->v.string[0] == '@') {
		expr = find_expression(value->v.string+1);
		if (!expr) {
			grecs_error(locus, 0,
				    _("expression %s was not defined"),
				    value->v.string + 1);
			return 1;
		}
		copy_expression(expr, pexpr);
	} else if (compile_expression(value->v.string, locus, &expr))
		return 1;
	else
		*pexpr = expr;
	
	return 0;
}

int
cb_server_variable(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	struct slb_varinst_tab *vt = varptr;
	struct slb_varinstance *inst;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	inst = varinst_lookupz(vt, value->v.arg.v[0].v.string, 1);
	if (!inst) {
		grecs_error(locus, 0, _("too many variable instances defined"));
		return 1;
	}
	if (inst->vi_name)
		grecs_error(locus, 0,
			     _("redefinition of %s"), inst->vi_name);
	
	inst->vi_oidlen = MAX_OID_LEN;
	if (!read_objid(value->v.arg.v[1].v.string,
			inst->vi_oid, &inst->vi_oidlen)) {
		grecs_error(locus, 0, _("cannot parse oid"));
		return 1;
	}

	inst->vi_name = value->v.arg.v[0].v.string;
	inst->vi_mib = value->v.arg.v[1].v.string;

	return 0;
}

int
cb_server_constant(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	struct slb_varinst_tab *vt = varptr;
	struct slb_varinstance *inst;
	char *p;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	inst = varinst_lookupz(vt, value->v.arg.v[0].v.string, 1);
	if (!inst) {
		grecs_error(locus, 0, _("too many constants defined"));
		return 1;
	}
	if (inst->vi_name)
		grecs_error(locus, 0,
			     _("redefinition of %s"), inst->vi_name);
	
	inst->vi_value = strtod(value->v.arg.v[1].v.string, &p);
	if (*p) {
		grecs_error(locus, 0, _("not a valid floating point number"));
		return 1;
	}
	
	inst->vi_name = value->v.arg.v[0].v.string;
	inst->vi_flags = SLV_VIF_CONST | SLV_VIF_SET;
	return 0;
}

int
cb_server_macro(enum grecs_callback_command cmd,
		grecs_locus_t *locus,
		void *varptr,
		grecs_value_t *value,
		void *cb_data)
{
	struct slb_macro_tab *mt = varptr;
	struct slb_macro *mac;
	
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_ARRAY || value->v.arg.c != 2) {
		grecs_error(locus, 0, _("expected two arguments"));
		return 1;
	}

	if (value->v.arg.v[0].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("first argument not a string"));
		return 1;
	}

	if (value->v.arg.v[1].type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("second argument not a string"));
		return 1;
	}

	mac = macro_lookupz(mt, value->v.arg.v[0].v.string, 1);
	if (!mac) {
		grecs_error(locus, 0, _("too many macros defined"));
		return 1;
	}
	if (mac->mac_name)
		grecs_error(locus, 0,
			    _("redefinition of macro %s"), mac->mac_name);
	
	mac->mac_name = value->v.arg.v[0].v.string;
	mac->mac_exp = value->v.arg.v[1].v.string;

	return 0;
}

static struct grecs_keyword server_kw[] = {
	{ "host", NULL, N_("Hostname or IP address"),
	  grecs_type_string, NULL, offsetof(struct slb_server, peer), NULL },
	{ "timeout", NULL, N_("Session timeout"),
	  grecs_type_time, NULL, offsetof(struct slb_server, timeout), NULL },
	{ "retries", NULL, N_("Number of retries before timeout"),
	  grecs_type_uint, NULL, offsetof(struct slb_server, retries), NULL },
	{ "community", NULL, N_("Community string"),
	  grecs_type_string, NULL, offsetof(struct slb_server, community),
	  NULL },
	{ "expression", NULL, N_("Evaluate expression"),
	  grecs_type_string, NULL, offsetof(struct slb_server, expr),
	  cb_server_expression },
	{ "variable", N_("<name: string> <oid: string>"),
	  N_("Define SNMP variable"),
	  grecs_type_string, NULL, offsetof(struct slb_server, varinst),
	  cb_server_variable },
	{ "constant", N_("<name: string> <value: string>"),
	  N_("Define a constant"),
	  grecs_type_string, NULL, offsetof(struct slb_server, varinst),
	  cb_server_constant },
	{ "macro", N_("<name: string> <expansion: string>"),
	  N_("Define a macro"),
	  grecs_type_string, NULL, offsetof(struct slb_server, macros),
	  cb_server_macro },
	{ NULL }
};

static int
cb_server(enum grecs_callback_command cmd,
	  grecs_locus_t *locus,
	  void *varptr,
	  grecs_value_t *value,
	  void *cb_data)
{
	struct slb_server *srv;
	void **pdata = cb_data;

	switch (cmd) {
	case grecs_callback_section_begin:
		if (!value || value->type != GRECS_TYPE_STRING) {
			grecs_error(locus, 0, _("tag must be a string"));
			return 0;
		}
		srv = xzalloc(sizeof(*srv));
		srv->id = xstrdup(value->v.string);
		*pdata = srv;
		break;

	case grecs_callback_section_end:
		srv = *pdata;
		if (!srv->peer) {
			grecs_error(locus, 0,
				    _("server name not defined"));
			return 1;
		}
		if (!srv->expr) {
			grecs_error(locus, 0,
				    _("server expression not defined"));
			return 1;
		}
		register_server(srv);
		*pdata = NULL;
		break;

	case grecs_callback_set_value:
		grecs_error(locus, 0, _("invalid use of block statement"));
	}
	return 0;
}


static int
cb_syslog_facility(enum grecs_callback_command cmd,
		   grecs_locus_t *locus,
		   void *varptr,
		   grecs_value_t *value,
		   void *cb_data)
{
	if (cmd != grecs_callback_set_value) {
		grecs_error(locus, 0, _("Unexpected block statement"));
		return 1;
	}
	if (!value || value->type != GRECS_TYPE_STRING) {
		grecs_error(locus, 0, _("expected string argument"));
		return 1;
	}

	if (string_to_syslog_facility(value->v.string, varptr))
		grecs_error(locus, 0, _("Unknown syslog facility `%s'"),
			    value->v.string);
	return 0;
}

static struct grecs_keyword syslog_kw[] = {
	{ "facility",
	  N_("name"),
	  N_("Set syslog facility. Arg is one of the following: user, daemon, "
	     "auth, authpriv, mail, cron, local0 through local7 (case-insensitive), "
	     "or a facility number."),
	  grecs_type_string, &syslog_facility, 0, cb_syslog_facility },
	{ "tag", N_("string"), N_("Tag syslog messages with this string"),
	  grecs_type_string, &syslog_tag },
	{ "print-priority", N_("arg"),
	  N_("Prefix each message with its priority"),
	  grecs_type_bool, &syslog_include_prio },
	{ NULL },
};


static struct grecs_keyword slb_kw[] = {
	{ "daemon", NULL, N_("Enable daemon mode"),
	  grecs_type_bool, &daemon_mode },
	{ "foreground", NULL, N_("Start in foreground even in daemon mode"),
	  grecs_type_bool, &foreground },
	{ "pidfile", N_("file"), N_("Set pid file name"),
	  grecs_type_string, &pidfile },
#if 0	
	{ "user", N_("name"), N_("Run with UID and GID of this user"),
	  grecs_type_string, NULL, 0, cb_user },
	{ "group", NULL, N_("Retain these supplementary groups"),
	  grecs_type_string|GRECS_LIST, NULL, 0, cb_supp_groups },
#endif

	{ "wakeup", NULL, N_("Wake up each <arg> seconds"),
	  grecs_type_time, &wakeup_interval },
	{ "skip", NULL, N_("Skip <arg> first measurements"),
	  grecs_type_uint, &snmp_skip_count },
	
	{ "mib-directory", N_("name"), N_("Search for MIBs in this directory"),
	  grecs_type_string, NULL, 0, cb_mib_directory },
	{ "add-mib", N_("file"), N_("Read MIBs from file"),
	  grecs_type_string|GRECS_LIST, &mib_file_list },

	{ "expression",
	  N_("<name: string> <expr: text>"),
	  N_("define a named expression"),
	  grecs_type_string, NULL, 0, cb_define_expression },

	{ "server", N_("id: string"), N_("Define a server"),
	  grecs_type_section, NULL, 0, cb_server, NULL, server_kw },

	{ "syslog", NULL, N_("Configure syslog logging"),
	  grecs_type_section, NULL, 0, NULL, NULL, syslog_kw },

	{ "tail", NULL,
	  N_("Output at most <arg> records from the sorted list"),
	  grecs_type_size, &output_tail_count },
	{ "head", NULL,
	  N_("Output at most <arg> records from the sorted list"),
	  grecs_type_size, &output_head_count },
	{ "output-file", NULL,
	  N_("Name of the output file or command"),
	  grecs_type_string, &output_file },
	{ "output-format", NULL,
	  N_("format string for output records"),
	  grecs_type_string, &output_format },
	{ "begin-output-message", NULL,
	  N_("string emitted before formatting output records"),
	  grecs_type_string, &begin_output_message },
	{ "end-output-message", NULL,
	  N_("string emitted after formatting output records"),
	  grecs_type_string, &end_output_message },
	
	{ NULL }
};

void
config_help()
{
	static char docstring[] =
		N_("Configuration file structure for SLB.\n"
		   "For more information, use `info slb configuration'.");
	grecs_format_docstring (stdout, docstring, 0);
	grecs_format_statement_array (stdout, slb_kw, 1, 0);
}

void
config_init()
{
	grecs_set_keywords (slb_kw);
#if 0
	grecs_include_path_setup (DEFAULT_VERSION_INCLUDE_DIR,
				  DEFAULT_INCLUDE_DIR, NULL);
#endif
	grecs_preprocessor = DEFAULT_PREPROCESSOR;
	grecs_log_to_stderr = true;
}

void
config_finish()
{
	struct slb_server *srv;
	gl_list_iterator_t itr;
	
	init_mib();

	/* Load requested MIBs */
	if (mib_file_list) {
		const void *p;
		
		itr = gl_list_iterator(mib_file_list);
	
		while (gl_list_iterator_next(&itr, &p, NULL)) {
			read_mib((const char*)p);
		}
		gl_list_iterator_free(&itr);
	}

	collect_output_refs();

	for (srv = srv_head; srv; srv = srv->next) {
		if (fixup_expression(srv->expr, &srv->varinst))
			srv->flags |= SLB_SRV_DISABLED;
	}
}	
