/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sysexits.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>

#include <regex.h>
#include <stdbool.h>
#include <stdio.h>
#include "c-ctype.h"
#include "obstack.h"
#include "progname.h"
#include "xalloc.h"
#include <inttypes.h>
#include "inttostr.h"
#include "hash.h"

#include "gettext.h"

#define _(s) gettext(s)
#define N_(s) s

#include "grecs.h"

#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>

extern int slb_sigtab[];
extern size_t slb_sigcount;
extern unsigned long slb_loop_serial;

void set_signals(void (*handler) (int signo), int *sigv, int sigc);

void config_help(void);
void config_init(void);
void config_finish(void);

extern int daemon_mode;
extern int foreground;
extern char *pidfile;
extern time_t wakeup_interval;


extern int debug_level;

#define debug(lev, s)					\
	do {						\
		if (debug_level >= (lev))		\
		       debug_printf s;			\
	} while(0)


enum node_type {
	node_unknown,
	node_variable,
	node_const,
	node_uminus,
	node_add,
	node_sub,
	node_mul,
	node_div,
	node_or,
	node_and,
	node_not,
	node_eq,
	node_ne,
	node_lt,
	node_le,
	node_gt,
	node_ge,
	node_cond,
	node_func
};

/* Maximum number of arguments to a built-in function */
#define SLB_MAX_ARGS 256

struct slb_node;
struct slb_expression;

typedef int (*slb_func_t) (struct slb_expression *exp,
			   struct slb_node *node,
			   int nargs, double *args,
			   double *res);

struct function {
	const char *name;
	slb_func_t func;
	int minargs;
	int maxargs;
};

struct node_state {
	struct node_state *next;
	void *node;
	void *key;
	size_t size;
	void *data;
};

struct slb_node {
	enum node_type n_type;
	struct slb_node *n_next;
	union {
		struct slb_node *n_arg[3];
		double n_const;
		int n_var;
		struct {
			slb_func_t func;
			struct slb_node *args;
			size_t nargs;
		} n_func;
	} v;
};

struct slb_node_list {
	struct slb_node *head;
	struct slb_node *tail;
	size_t count;
};

struct slb_varent {
	char *ve_name;
	struct slb_varinstance *ve_inst;
};

struct slb_vartab {
	size_t vt_max;
	struct slb_varent vt_tab[SLB_MAX_ARGS];
};

#define SLV_VIF_CONST 0x01
#define SLV_VIF_USED  0x02
#define SLV_VIF_SET   0x04

struct slb_varinstance {
	const char *vi_name;
	const char *vi_mib;
	oid vi_oid[MAX_OID_LEN];
	size_t vi_oidlen;
	double vi_value;
	int vi_flags;
};

struct slb_varinst_tab {
	size_t vit_max;
	struct slb_varinstance vit_tab[SLB_MAX_ARGS];
};
	
struct slb_expression {
	const char *ex_name;
	const char *ex_text;
	grecs_locus_t ex_locus;
	struct slb_vartab ex_vartab;
	struct node_state *state;
	struct slb_node *ex_tree;
};

struct slb_macro {
	const char *mac_name;
	const char *mac_exp;
};

struct slb_macro_tab {
	size_t mt_max;
	struct slb_macro mt_tab[SLB_MAX_ARGS];
};

int register_expression(struct slb_expression *expr);
struct slb_expression *find_expression(const char *name);
void copy_expression(struct slb_expression *src, struct slb_expression **dst);
int fixup_expression(struct slb_expression *expr, struct slb_varinst_tab *vit);

int compile_expression(const char *text, grecs_locus_t *locus,
		       struct slb_expression **pexpr);
int eval_expression(struct slb_expression *expr, double *pnum);

int vartab_lookup(struct slb_vartab *vt, const char *ident, size_t len);
int vartab_install(struct slb_vartab *vt, const char *ident, size_t len);

struct slb_varinstance *varinst_lookup(struct slb_varinst_tab *vit,
				       const char *name, size_t len,
				       int install);
struct slb_varinstance *varinst_lookupz(struct slb_varinst_tab *vit,
					const char *name,
					int install);
struct slb_varinstance *varinst_lookup_oid(struct slb_varinst_tab *vit,
					   oid *oid, size_t len);
void varinst_clear(struct slb_varinst_tab *vit);
int varinst_allset(struct slb_varinst_tab *vit);

struct slb_macro *macro_lookup(struct slb_macro_tab *mt,
			       const char *name, size_t len,
			       int install);
struct slb_macro *macro_lookupz(struct slb_macro_tab *mt, const char *name,
				int install);

#define SLB_SRV_DISABLED 0x01
#define SLB_SRV_COMPUTED 0x02

struct slb_server {
	char *id;
	char *peer;
	int port;
	time_t timeout;
	unsigned retries;
	char *community;
	struct slb_varinst_tab varinst;
	struct slb_expression *expr;
	struct snmp_session *sess;
	double weight;
	int flags;
	struct slb_macro_tab macros;
	struct slb_server *next;
};

extern struct slb_server *srv_head, *srv_tail;
extern size_t srv_count;
extern struct slb_server **srvtab;
extern size_t srvtab_count;
extern unsigned snmp_skip_count;

void snmploop(void);


extern int log_to_stderr;
extern int syslog_facility;
extern const char *syslog_tag;
extern int syslog_include_prio;

void logger_setup(void);
void logmsg(int prio, const char *fmt, ...) SLB_PRINTFLIKE(2,3);
void logmsg_at_locus(int prio, grecs_locus_t *locus, const char *fmt, ...)
	SLB_PRINTFLIKE(3,4);
void debug_printf(const char *fmt, ...) SLB_PRINTFLIKE(1,2);

int string_to_syslog_facility(const char *str, int *pfacility);


extern size_t output_tail_count;
extern size_t output_head_count;
extern char *output_file;
extern char *output_format;
extern char *begin_output_message;
extern char *end_output_message;

void slb_set_output_defaults(void);
void slb_open_output(void);
void slb_close_output(void);
void slb_check_output(void);

void slb_fmtout(void);

void collect_output_refs(void);
int is_output_refname(const char *name);

void close_fds(fd_set *fdset);


